import argparse
import sys
from pathlib import Path

from taihe.driver.backend import BackendRegistry
from taihe.driver.contexts import CompilerInstance, CompilerInvocation
from taihe.utils.build_metadata import BuildMetadata
from taihe.utils.outputs import CMakeOutputConfig, OutputConfig
from taihe.utils.resources import (
    ResourceContext,
    RuntimeHeader,
    RuntimeSource,
)


def main():
    registry = BackendRegistry()
    registry.register_all()

    parser = argparse.ArgumentParser(
        prog="taihec",
        description="generates source code from taihe files",
    )
    parser.add_argument(
        "src_files",
        type=Path,
        nargs="*",
        default=[],
        help="input .taihe files, if not provided, will read from stdin",
    )
    parser.add_argument(
        "-I",
        type=Path,
        dest="src_dirs",
        nargs="*",
        default=[],
        help="directories of .taihe source files",
    )  # deprecated
    parser.add_argument(
        "--output",
        "-O",
        type=Path,
        dest="dst_dir",
        default="taihe-generated",
        help="directory for generated files",
    )
    parser.add_argument(
        "--generate",
        "-G",
        dest="backends",
        nargs="*",
        default=[],
        choices=registry.get_backend_names(),
        help="backends to generate sources, default: abi-header, abi-source, c-author",
    )
    parser.add_argument(
        "--codegen",
        "-C",
        dest="config",
        action="append",
        default=[],
        help="additional code generation configuration",
    )
    parser.add_argument(
        "--build",
        "-B",
        dest="build_system",
        choices=["cmake"],
        help="build system to use for generated sources",
    )

    # Special options {{
    ResourceContext.register_cli_options(parser)
    parser.add_argument("--version", action="store_true")

    args = parser.parse_args()
    if args.version:
        BuildMetadata.get().print_info(tool="Taihe compiler (taihec)", auto_exit=True)
    ResourceContext.initialize(args)
    # }} Special options

    backends = registry.collect_required_backends(args.backends)
    resolved_backends = [b() for b in backends]

    if args.build_system == "cmake":
        output_config = CMakeOutputConfig(
            dst_dir=Path(args.dst_dir),
            runtime_include_dir=RuntimeHeader.resolve_path(),
            runtime_src_dir=RuntimeSource.resolve_path(),
        )
    else:
        output_config = OutputConfig(
            dst_dir=Path(args.dst_dir),
        )

    if not args.src_files and not args.src_dirs:
        print("taihec: error: no input files", file=sys.stderr)
        return -1

    invocation = CompilerInvocation(
        src_files=args.src_files,
        src_dirs=args.src_dirs,
        output_config=output_config,
        backends=resolved_backends,
    )

    for config in args.config:
        k, *v = config.split("=", 1)
        if k == "sts:keep-name":
            invocation.sts_keep_name = True
        elif k == "arkts:module-prefix":
            invocation.arkts_module_prefix = v[0] if v else None
        elif k == "arkts:path-prefix":
            invocation.arkts_path_prefix = v[0] if v else None
        else:
            raise ValueError(f"unknown codegen config {k!r}")

    instance = CompilerInstance(invocation)

    if not instance.run():
        return -1
    return 0


if __name__ == "__main__":
    sys.exit(main())
